"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _cicadaJsLib = require("cicada-js-lib");

var _cicadaJsLib2 = _interopRequireDefault(_cicadaJsLib);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var is = _cicadaJsLib2.default.is;
var BUFFER = _cicadaJsLib2.default.BUFFER;
var STACK = _cicadaJsLib2.default.STACK;
var HASH_TABLE = _cicadaJsLib2.default.HASH_TABLE;

function string_to_sum(string) {
  var sum = 0;
  var max_step = 10;
  for (var i = 0; i < string.length; i++) {
    sum = sum + string.codePointAt(i) * (2 << Math.min(i, max_step));
  }
  return sum;
}

var tag_hash_table = new HASH_TABLE(
// prime table size
997,
// key_equal
function (key1, key2) {
  return key1 === key2;
},
// hash
function (key, counter) {
  return (counter + string_to_sum(key)) % 997;
});

function array_equal(a, b) {
  if (a === b) return true;
  if (a == null || b == null) return false;
  if (a.length != b.length) return false;
  for (var i = 0; i < a.length; ++i) {
    if (a[i] !== b[i]) return false;
  }
  return true;
}

Array.prototype.last = function () {
  return this[this.length - 1];
};

function array_remove_dup(orig_array) {
  var array = orig_array.map(function (element) {
    return element;
  });
  array.sort();
  var result = [];
  for (var i = 0; i < array.length; i++) {
    if (result.last() === array[i]) {} else {
      result.push(array[i]);
    }
  }
  return result;
}

function array_remove(array, value) {
  var result = [];
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = array[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var element = _step.value;

      if (element === value) {
        // nothing
      } else {
          result.push(element);
        }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return result;
}

var name_hash_table = new HASH_TABLE(
// prime table size
997,
// key_equal
function (key1, key2) {
  return array_equal(array_remove_dup(key1), array_remove_dup(key2));
},
// hash
function (key, counter) {
  var sum = array_remove_dup(key).map(string_to_sum).reduce(function (sum, v) {
    return sum + v;
  });
  return (counter + sum) % 997;
});

function string_to_key(string) {
  return array_remove(string.split(":"), "");
}

function key_to_string(key) {
  var result = "";
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = key[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var element = _step2.value;

      result = result.concat(":", element);
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  return result.substring(1);
}

function name(string) {
  return name_hash_table.key_to_index(string_to_key(string));
}

function name_to_string(name) {
  return key_to_string(name_hash_table.index_to_entry(name).key);
}

function set(name, value) {
  name_hash_table.index_to_entry(name).value = value;
}

function get(name) {
  var value = name_hash_table.index_to_entry(name).value;
  if (value === null) {
    // better error handling
    console.log("- get");
    console.log("  no value found from name:", name_to_string(name));
  } else {
    return value;
  }
}

function G(typename, type, data) {
  this.typename = typename;
  this.type = type;
  this.data = data;
}

G.prototype = {};

function L(typename, type, data) {
  G.call(this, typename, type, data);
  this.determinacy = 0;
}

L.prototype = {};

var argument_stack = new STACK(1024, 0);

function A(typename, data) {
  this.typename = typename;
  this.data = data;
}

var return_stack = new STACK(1024, 0);

function R(array) {
  this.cursor = 0;
  this.array = array;
  this.local_variable_map = new Map();
}

R.prototype = {
  get_current_jo: function get_current_jo() {
    return this.array[this.cursor];
  },
  next: function next() {
    this.cursor = 1 + this.cursor;
  }
};

function JO(explainer, data) {
  this.explainer = explainer;
  this.data = data;
}

JO.prototype = {
  happy: function happy() {
    this.explainer(this.data);
  }
};

function primitive_function_explainer(name) {
  get(name).data.call();
};

function function_explainer(name) {
  return_stack.push(new R(get(name).data));
};

function variable_explainer(name) {
  argument_stack.push(new A(get(name).typename, get(name).data));
};

return_stack.eval = function () {
  try {
    while (true) {
      var r = return_stack.tos();
      var jo = r.get_current_jo();
      r.next();
      jo.happy();
    }
  } catch (string) {
    if (string === "bye") {
      return;
    } else {
      console.log("- unknown exception:");
      console.log("  ", string);
      return;
    }
  }
};

function def(name_string, typename_string, type, data) {
  set(name(name_string), new G(name(typename_string), type, data));
}

def("end", "->", null, function () {
  return_stack.pop();
});

def("bye", "->", null, function () {
  console.log("bye bye ^-^/");
  throw "bye";
});

def("dup", "->", null, function () {
  var a = argument_stack.tos();
  argument_stack.push(a);
});

def("mul", "->", null, function () {
  var a1 = argument_stack.pop();
  var a2 = argument_stack.pop();
  argument_stack.push(new A(name("integer"), a1.data * a2.data));
});

def("simple-wirte", "->", null, function () {
  var a = argument_stack.pop();
  console.log(name_to_string(a.typename), a.data);
});

def("little-test-number", "integer", "integer", 4);

def("square", "->", null, [new JO(primitive_function_explainer, name("dup")), new JO(primitive_function_explainer, name("mul")), new JO(primitive_function_explainer, name("end"))]);

def("little-test", "->", null, [new JO(variable_explainer, name("little-test-number")), new JO(function_explainer, name("square")), new JO(primitive_function_explainer, name("simple-wirte")), new JO(primitive_function_explainer, name("bye"))]);

function begin() {
  return_stack.push(new R([new JO(function_explainer, name("little-test")), new JO(primitive_function_explainer, name("end"))]));
  return_stack.eval();
}

begin();

exports.default = {
  tag_hash_table: tag_hash_table, name_hash_table: name_hash_table, name: name, set: set, get: get
};