import lib from 'cicada-js-lib';
const {
  is,
  BUFFER, STACK,
  HASH_TABLE,
} = lib;

function string_to_sum (string) {
  let sum = 0;
  let max_step = 10;
  for (let i = 0; i < string.length; i++) {
    sum = sum +
      string.codePointAt(i) *
      (2 << Math.min(i, max_step));
  }
  return sum;
}

const tag_hash_table = new HASH_TABLE(
  // prime table size
  997,
  // key_equal
  (key1, key2) => {
    return key1 === key2;
  },
  // hash
  (key, counter) => {
    return (counter + string_to_sum(key)) % 997;
  }
);

function array_equal (a, b) {
  if (a === b) return true;
  if (a == null || b == null) return false;
  if (a.length != b.length) return false;
  for (var i = 0; i < a.length; ++i) {
    if (a[i] !== b[i]) return false;
  }
  return true;
}

Array.prototype.last = function () {
  return this[this.length - 1];
};

function array_remove_dup (orig_array) {
  const array = orig_array.map((element) => {return element;});
  array.sort();
  let result = [];
  for (let i = 0; i < array.length; i++) {
    if (result.last() === array[i]) {
    } else {
      result.push(array[i]);
    }
  }
  return result;
}

function array_remove (array, value) {
  let result = [];
  for (let element of array) {
    if (element === value) {
      // nothing
    } else {
      result.push(element);
    }
  }
  return result;
}

const name_hash_table = new HASH_TABLE(
  // prime table size
  997,
  // key_equal
  (key1, key2) => {
    return array_equal(array_remove_dup(key1),
                       array_remove_dup(key2));
  },
  // hash
  (key, counter) => {
    let sum =
        array_remove_dup(key)
        .map(string_to_sum)
        .reduce((sum, v) => {return sum + v;});
    return (counter + sum) % 997;
  }
);

function string_to_key (string) {
  return array_remove(string.split(":"), "");
}

function key_to_string (key) {
  let result = "";
  for (let element of key) {
    result = result.concat(":", element);
  }
  return result.substring(1);
}

function name (string) {
  return name_hash_table.key_to_index(string_to_key(string));
}

function name_to_string (name) {
  return key_to_string(name_hash_table.index_to_entry(name).key);
}

function set (name, value) {
  name_hash_table.index_to_entry(name).value = value;
}

function get (name) {
  let value = name_hash_table.index_to_entry(name).value;
  if (value === null) {
    // better error handling
    console.log("- get");
    console.log("  no value found from name:", name_to_string(name));
  } else {
    return value;
  }
}

function G (typename, type, data) {
  this.typename = typename;
  this.type = type;
  this.data = data;
}

G.prototype = {
};

function L (typename, type, data) {
  G.call(this, typename, type, data);
  this.determinacy = 0;
}

L.prototype = {
};

const argument_stack = new STACK(1024, 0);

function A (typename, data) {
  this.typename = typename;
  this.data = data;
}

const return_stack = new STACK(1024, 0);

function R (array) {
  this.cursor = 0;
  this.array = array;
  this.local_variable_map = new Map();
}

R.prototype = {
  get_current_jo: function () {
    return this.array[this.cursor];
  },
  next: function () {
    this.cursor = 1 + this.cursor;
  }
};

function JO (explainer, data) {
  this.explainer = explainer;
  this.data = data;
}

JO.prototype = {
  happy: function () {
    this.explainer(this.data);
  }
};

function primitive_function_explainer (name) {
  get(name).data.call();
};

function function_explainer (name) {
  return_stack.push(
    new R (get(name).data));
};

function variable_explainer (name) {
  argument_stack.push(
    new A (get(name).typename,
           get(name).data));
};

return_stack.eval = function () {
  try {
    while (true) {
      const r = return_stack.tos();
      const jo = r.get_current_jo();
      r.next();
      jo.happy();
    }
  } catch (string) {
    if (string === "bye") {
      return;
    } else {
      console.log("- unknown exception:");
      console.log("  ", string);
      return;
    }
  }
};

function def (name_string, typename_string, type, data) {
  set(name(name_string),
      new G (name(typename_string),
             type,
             data));
}

def ("end", "->", null, () => {
  return_stack.pop();
});

def ("bye", "->", null, () => {
  console.log("bye bye ^-^/");
  throw "bye";
});

def ("dup", "->", null, () => {
  let a = argument_stack.tos();
  argument_stack.push(a);
});

def ("mul", "->", null, () => {
  let a1 = argument_stack.pop();
  let a2 = argument_stack.pop();
  argument_stack
    .push(new A (
      name("integer"),
      a1.data * a2.data));
});

def ("simple-wirte", "->", null, () => {
  let a = argument_stack.pop();
  console.log(name_to_string(a.typename),
              a.data);
});

def ("little-test-number", "integer", "integer", 4);

def ("square", "->", null,
     [ new JO (primitive_function_explainer, name("dup")),
       new JO (primitive_function_explainer, name("mul")),
       new JO (primitive_function_explainer, name("end")),
     ]
    );

def("little-test", "->", null,
    [ new JO (variable_explainer, name("little-test-number")),
      new JO (function_explainer, name("square")),
      new JO (primitive_function_explainer, name("simple-wirte")),
      new JO (primitive_function_explainer, name("bye")),
    ]
   );

function begin () {
  return_stack.push(new R ([
    new JO (function_explainer, name("little-test")),
    new JO (primitive_function_explainer, name("end")),
  ]));
  return_stack.eval();
}

begin();

export default {
  tag_hash_table, name_hash_table, name, set, get,
};
